@@ -92,15 +92,22 @@ function HastexoXBlock(runtime, element, configuration) {
         }
 
         /* Process terminal URL. */
-        var prot_map = {
-            "http:":  "ws:",
-            "https:": "wss:"
-        };
-        var terminal_http_url = location.protocol + '//' + location.hostname + configuration.terminal_url;
-        var terminal_ws_url = prot_map[location.protocol] + '//' + location.hostname + configuration.terminal_url;
+        var terminal_http_url;
+        var terminal_ws_url;
+        if (configuration.terminal_url.startsWith('http')) {
+          terminal_http_url = configuration.terminal_url;
+          terminal_ws_url = configuration.terminal_url.replace('http', 'ws');
+        } else {
+          var prot_map = {
+              "http:":  "ws:",
+              "https:": "wss:"
+          };
+          terminal_http_url = location.protocol + '//' + location.hostname + configuration.terminal_url;
+          terminal_ws_url = prot_map[location.protocol] + '//' + location.hostname + configuration.terminal_url;
+        }
 
         /* Load app dynamically. */
-        $.cachedScript(terminal_http_url + '/guacamole-common-js/all.min.js').done(function() {
+        $.cachedScript(terminal_http_url + 'guacamole-common-js/all.min.js').done(function() {
             terminal_client = new Guacamole.Client(
                 new Guacamole.WebSocketTunnel(terminal_ws_url + "websocket-tunnel")
             );
@@ -373,6 +380,25 @@ function HastexoXBlock(runtime, element, configuration) {
                 get_user_stack_status,
                 fuzz_timeout(configuration.timeouts['status'])
             );
+        } else if (stack.status == 'SUSPEND_PENDING' || stack.status == 'DELETE_PENDING') {
+            /* Stack is pending.  Display retry message. */
+            if (keepalive_timer) clearTimeout(keepalive_timer);
+            if (idle_timer) clearTimeout(idle_timer);
+            var dialog = $('#launch_error');
+            var dialog_msg = "Your lab environment is undergoing maintenance";
+            var error_msg =
+                "Your lab environment is undergoing automatic maintenance. " +
+                "Please try again in a few minutes.";
+            dialog.find('.message').html(dialog_msg);
+            dialog.find('.error_msg').html(error_msg);
+            dialog.find('input.ok').one('click', function() {
+                $.dialog.close();
+            });
+            dialog.find('input.retry').one('click', function() {
+                $.dialog.close();
+                location.reload();
+            });
+            dialog.dialog(element);
         } else {
             /* Unexpected status.  Display error message. */
             if (keepalive_timer) clearTimeout(keepalive_timer);
