@@ -1,22 +1,38 @@
 import time
 import os
-import uuid
-import paramiko
 import traceback
 import socket
 
-from django.db import transaction
+from django.db import transaction, close_old_connections
+from django.db.utils import OperationalError
 from celery import Task
 from celery.utils.log import get_task_logger
 from celery.exceptions import SoftTimeLimitExceeded
-from io import StringIO
-from paramiko.ssh_exception import (AuthenticationException,
-                                    SSHException,
-                                    NoValidConnectionsError)
 
 from .models import Stack
 from .provider import Provider, ProviderException
-from .common import (OCCUPANCY_STATES, get_xblock_settings, update_stack)
+from .common import (
+    DELETE,
+    IN_PROGRESS,
+    UP_STATES,
+    OCCUPANCY_STATES,
+    RESUME_COMPLETE,
+    RESUME_FAILED,
+    SUSPEND_COMPLETE,
+    SUSPEND_FAILED,
+    CREATE_FAILED,
+    DELETE_COMPLETE,
+    DELETE_IN_PROGRESS,
+    DELETE_FAILED,
+    LAUNCH_TIMEOUT,
+    get_xblock_settings,
+    update_stack_fields,
+    ssh_to,
+    read_from_contentstore,
+    remote_exec,
+    RemoteExecException,
+    RemoteExecTimeout,
+)
 
 logger = get_task_logger(__name__)
 
@@ -46,7 +62,87 @@ class LaunchStackFailed(Exception):
             self.delete = True
 
 
-class LaunchStackTask(Task):
+class HastexoTask(Task):
+    """
+    Abstract task.
+
+    """
+    sleep_timeout = None
+
+    def get_sleep_timeout(self):
+        if self.sleep_timeout is None:
+            settings = get_xblock_settings()
+            self.sleep_timeout = settings.get("sleep_timeout", 10)
+
+        return self.sleep_timeout
+
+    def sleep(self, multiplier=1):
+        sleep_time = multiplier * self.get_sleep_timeout()
+        logger.debug("Sleeping for %i seconds" % sleep_time)
+        time.sleep(sleep_time)
+
+    @transaction.atomic
+    def update_stack(self, data):
+        stack = Stack.objects.select_for_update().get(id=self.stack_id)
+        update_stack_fields(stack, data)
+        stack.save(update_fields=list(data.keys()))
+
+    def update_stack_retry(self, data, max_attempts=3):
+        """A wrapper around update_stack that retries until it either
+        succeeds, or times out."""
+
+        # This wraps a try/except block *around* update_stack, which
+        # uses transaction.atomic().
+        # Catching exceptions *inside* transaction.atomic is highly
+        # discouraged.
+        #
+        # Reference:
+        # https://docs.djangoproject.com/en/1.11/topics/db/transactions/
+        i = 1
+        while True:
+            try:
+                # Try updating the stack. If it succeeds, we're done.
+                self.update_stack(data)
+                break
+            except SoftTimeLimitExceeded:
+                # If we've hit the timeout, something went massively
+                # wrong, repeatedly. Log an error, and throw the
+                # timeout up the stack.
+                logger.error("Timeout updating database status for "
+                             "[%s]." % self.stack_name)
+                raise
+            except OperationalError:
+                # Sleep and then retry, until we either succeed, hit
+                # max_attempts, or hit the timeout. Between attempts,
+                # increase the sleep period by factors of 1, 2, 3.
+
+                # First, clean up the broken connection. Django will
+                # subsequently reconnect as needed.
+                close_old_connections()
+
+                if i >= max_attempts:
+                    logger.error(
+                        "Unable to update database status for "
+                        "[%s] (%i attempts)." % (self.stack_name, i)
+                    )
+                    raise
+                else:
+                    logger.warning(
+                        "Error updating database status for "
+                        "[%s], retrying (attempt %i)" % (self.stack_name, i),
+                        exc_info=True
+                    )
+                    self.sleep(multiplier=i)
+                    i += 1
+
+        # If we're here, we've successfully updated the database
+        # stack. This is the normal flow of events, so we only need to
+        # log this when we're debugging.
+        logger.debug("Updated database status for [%s] " %
+                     self.stack_name)
+
+
+class LaunchStackTask(HastexoTask):
     """
     Launch, or if it already exists and is suspended, resume a stack for the
     user.
@@ -58,32 +154,56 @@ class LaunchStackTask(Task):
         Run the celery task.
 
         """
+        # Set the arguments.
+        for key, value in kwargs.items():
+            setattr(self, key, value)
+
+        # Get the stack
+        stack = Stack.objects.get(id=self.stack_id)
+
+        # Initialize parameters
+        self.protocol = stack.protocol
+        self.port = stack.port
+        self.stack_run = stack.run
+        self.stack_name = stack.name
+        self.stack_user_name = stack.user
+        self.course_id = stack.course_id
+        self.student_id = stack.student_id
+        self.hook_events = stack.hook_events
+
+        # Initialize hook script
+        self.hook_script = None
+        if stack.hook_script:
+            self.hook_script = read_from_contentstore(
+                stack.course_id,
+                stack.hook_script
+            )
+
         # Initialize providers
         self.providers = []
-        for provider in kwargs.get("providers", []):
+        for provider in stack.providers:
             p = Provider.init(provider["name"])
             p.set_capacity(provider["capacity"])
-            p.set_template(provider["template"])
 
-            environment = provider.get("environment")
-            if environment:
+            template = read_from_contentstore(
+                stack.course_id,
+                provider["template"]
+            )
+            p.set_template(template)
+
+            environment_path = provider.get("environment")
+            if environment_path:
+                environment = read_from_contentstore(
+                    stack.course_id,
+                    environment_path
+                )
                 p.set_environment(environment)
 
             self.providers.append(p)
 
-        # Set the rest of the arguments.
-        for key, value in kwargs.items():
-            if key == "providers":
-                continue
-            setattr(self, key, value)
-
-        self.settings = get_xblock_settings()
-        task_timeouts = self.settings.get('task_timeouts', {})
-        self.sleep_seconds = task_timeouts.get('sleep', 5)
-
         try:
             # Launch the stack and wait for it to complete.
-            stack_data = self.launch_stack()
+            stack_data = self.launch_stack(stack.provider)
         except LaunchStackFailed as e:
             logger.error(e.error_msg)
 
@@ -106,16 +226,9 @@ class LaunchStackTask(Task):
             # Roll back in case of failure
             self.cleanup_stack(e)
 
-        return stack_data
-
-    def update_stack(self, data):
-        return update_stack(self.stack_name,
-                            self.course_id,
-                            self.student_id,
-                            data)
-
-    def sleep(self):
-        time.sleep(self.sleep_seconds)
+        # Don't wait for the user to check results.  Update the database
+        # immediately.
+        self.update_stack_retry(stack_data)
 
     def get_provider(self, name):
         try:
@@ -125,7 +238,7 @@ class LaunchStackTask(Task):
 
         return provider
 
-    def launch_stack(self):
+    def launch_stack(self, provider_name=None):
         """
         Launch the user stack, either by creating or resuming it.  If a reset
         is requested, delete the stack and recreate it.
@@ -133,17 +246,8 @@ class LaunchStackTask(Task):
         """
         logger.info("Launching stack [%s]." % self.stack_name)
 
-        # Fetch stack information from the database, but do so atomically to
-        # make sure the original request had a chance to commit.
-        with transaction.atomic():
-            stack = Stack.objects.select_for_update().get(
-                student_id=self.student_id,
-                course_id=self.course_id,
-                name=self.stack_name
-            )
-
-        if stack.provider:
-            provider = self.get_provider(stack.provider)
+        if provider_name:
+            provider = self.get_provider(provider_name)
             if self.reset:
                 self.try_provider(provider, True)
 
@@ -195,7 +299,7 @@ class LaunchStackTask(Task):
                 # If this is the last provider, or it was a timeout, re-raise
                 # the exception and let the parent clean up.
                 if (index == (len(self.providers) - 1) or
-                   e.status == "LAUNCH_TIMEOUT"):
+                   e.status == LAUNCH_TIMEOUT):
                     raise
                 else:
                     logger.error(e.error_msg)
@@ -206,7 +310,7 @@ class LaunchStackTask(Task):
                     except SoftTimeLimitExceeded:
                         error_msg = "Timeout cleaning up stack [%s]." % (
                             self.stack_name)
-                        raise LaunchStackFailed(e.provider, "LAUNCH_TIMEOUT",
+                        raise LaunchStackFailed(e.provider, LAUNCH_TIMEOUT,
                                                 error_msg)
 
         if stack_data is None:
@@ -214,7 +318,7 @@ class LaunchStackTask(Task):
                          (self.stack_name))
             error_msg = ("There are no providers available to launch your "
                          "environment in.")
-            raise LaunchStackFailed("", "CREATE_FAILED", error_msg)
+            raise LaunchStackFailed("", CREATE_FAILED, error_msg)
 
         return stack_data
 
@@ -241,16 +345,16 @@ class LaunchStackTask(Task):
         except ProviderException as e:
             error_msg = ("Error retrieving [%s] stack information: %s" %
                          (self.stack_name, e))
-            raise LaunchStackFailed(provider, "CREATE_FAILED", error_msg)
+            raise LaunchStackFailed(provider, CREATE_FAILED, error_msg)
         except SoftTimeLimitExceeded:
             error_msg = "Timeout fetching stack [%s] information." % (
                 self.stack_name)
-            raise LaunchStackFailed(provider, "LAUNCH_TIMEOUT", error_msg)
+            raise LaunchStackFailed(provider, LAUNCH_TIMEOUT, error_msg)
 
         # If stack is undergoing a change of state, wait until it
         # finishes.
         try:
-            while 'IN_PROGRESS' in provider_stack["status"]:
+            while IN_PROGRESS in provider_stack["status"]:
                 try:
                     # Sleep to avoid throttling.
                     self.sleep()
@@ -259,17 +363,21 @@ class LaunchStackTask(Task):
                 except ProviderException as e:
                     error_msg = ("Error waiting for stack [%s] to change "
                                  "state: %s" % (self.stack_name, e))
-                    raise LaunchStackFailed(provider, "CREATE_FAILED",
-                                            error_msg)
+                    if DELETE in provider_stack["status"]:
+                        raise LaunchStackFailed(provider, CREATE_FAILED,
+                                                error_msg)
+                    else:
+                        raise LaunchStackFailed(provider, RESUME_FAILED,
+                                                error_msg, CLEANUP_SUSPEND)
         except SoftTimeLimitExceeded:
             error_msg = "Timeout waiting for stack [%s] state change." % (
                 self.stack_name)
-            raise LaunchStackFailed(provider, "LAUNCH_TIMEOUT", error_msg)
+            raise LaunchStackFailed(provider, LAUNCH_TIMEOUT, error_msg)
 
         # Reset the stack, if necessary
         if reset:
             try:
-                if provider_stack["status"] != 'DELETE_COMPLETE':
+                if provider_stack["status"] != DELETE_COMPLETE:
                     # Sleep to avoid throttling.
                     self.sleep()
 
@@ -278,14 +386,14 @@ class LaunchStackTask(Task):
             except ProviderException as e:
                 error_msg = ("Error deleting stack [%s]: %s" %
                              (self.stack_name, e))
-                raise LaunchStackFailed(provider, "CREATE_FAILED", error_msg)
+                raise LaunchStackFailed(provider, CREATE_FAILED, error_msg)
             except SoftTimeLimitExceeded:
                 error_msg = "Timeout resetting stack [%s]." % self.stack_name
-                raise LaunchStackFailed(provider, "LAUNCH_TIMEOUT", error_msg)
+                raise LaunchStackFailed(provider, LAUNCH_TIMEOUT, error_msg)
         else:
             # Create the stack if it doesn't exist
             try:
-                if provider_stack["status"] == 'DELETE_COMPLETE':
+                if provider_stack["status"] == DELETE_COMPLETE:
                     # Sleep to avoid throttling.
                     self.sleep()
 
@@ -295,20 +403,20 @@ class LaunchStackTask(Task):
             except ProviderException as e:
                 error_msg = ("Error creating stack [%s]: %s" %
                              (self.stack_name, e))
-                raise LaunchStackFailed(provider, "CREATE_FAILED", error_msg,
+                raise LaunchStackFailed(provider, CREATE_FAILED, error_msg,
                                         CLEANUP_DELETE)
             except SoftTimeLimitExceeded:
                 error_msg = "Timeout creating stack [%s]." % self.stack_name
-                raise LaunchStackFailed(provider, "LAUNCH_TIMEOUT", error_msg,
+                raise LaunchStackFailed(provider, LAUNCH_TIMEOUT, error_msg,
                                         CLEANUP_DELETE)
             # If stack is suspended, resume it.
             try:
-                if provider_stack["status"] == 'SUSPEND_COMPLETE' or \
-                        provider_stack["status"] == 'RESUME_COMPLETE':
+                if provider_stack["status"] == SUSPEND_COMPLETE or \
+                        provider_stack["status"] == RESUME_COMPLETE:
                     # Store the fact the stack was resumed
                     was_resumed = True
 
-                if provider_stack["status"] == 'SUSPEND_COMPLETE':
+                if provider_stack["status"] == SUSPEND_COMPLETE:
                     # Sleep to avoid throttling.
                     self.sleep()
 
@@ -317,11 +425,11 @@ class LaunchStackTask(Task):
             except ProviderException as e:
                 error_msg = ("Error resuming stack [%s]: %s" %
                              (self.stack_name, e))
-                raise LaunchStackFailed(provider, "RESUME_FAILED", error_msg,
+                raise LaunchStackFailed(provider, RESUME_FAILED, error_msg,
                                         CLEANUP_SUSPEND)
             except SoftTimeLimitExceeded:
                 error_msg = "Timeout resuming stack [%s]." % self.stack_name
-                raise LaunchStackFailed(provider, "LAUNCH_TIMEOUT", error_msg,
+                raise LaunchStackFailed(provider, LAUNCH_TIMEOUT, error_msg,
                                         CLEANUP_SUSPEND)
 
             # Launch completed successfully.  Wait for provisioning, collect
@@ -336,7 +444,7 @@ class LaunchStackTask(Task):
                     cleanup = CLEANUP_DELETE
 
                 error_msg = "Timeout verifying stack [%s]." % self.stack_name
-                raise LaunchStackFailed(provider, "LAUNCH_TIMEOUT", error_msg,
+                raise LaunchStackFailed(provider, LAUNCH_TIMEOUT, error_msg,
                                         cleanup)
 
             stack_data = {
@@ -377,48 +485,32 @@ class LaunchStackTask(Task):
                                      self.stack_name, e))
 
     def wait_for_ping(self, stack_ip):
-        ping_command = PING_COMMAND % (self.sleep_seconds, stack_ip)
+        ping_command = PING_COMMAND % (self.get_sleep_timeout(), stack_ip)
         while os.system(ping_command) != 0:
             self.sleep()
 
-    def setup_ssh(self, stack_key):
-        ssh = paramiko.SSHClient()
-        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
-        pkey = paramiko.RSAKey.from_private_key(StringIO(stack_key))
-
-        return (ssh, pkey)
-
     def wait_for_ssh(self, stack_key, stack_ip, was_resumed, provider):
-        ssh, pkey = self.setup_ssh(stack_key)
-        connected = False
-        while not connected:
-            try:
-                ssh.connect(stack_ip, username=self.stack_user_name, pkey=pkey)
-            except (EOFError,
-                    AuthenticationException,
-                    SSHException,
-                    NoValidConnectionsError):
-                self.sleep()
-            except SoftTimeLimitExceeded:
-                raise
-            except Exception:
-                if was_resumed:
-                    error_status = 'RESUME_FAILED'
-                    cleanup = CLEANUP_SUSPEND
-                else:
-                    error_status = 'CREATE_FAILED'
-                    cleanup = CLEANUP_DELETE
-
-                logger.error("Exception when checking SSH connection to stack "
-                             "[%s]: %s" % (self.stack_name,
-                                           traceback.format_exc()))
-                error_msg = ("Could not connect to your lab environment "
-                             "[%s]." % self.stack_name)
-                raise LaunchStackFailed(provider, error_status, error_msg,
-                                        cleanup)
+        try:
+            ssh = ssh_to(self.stack_user_name, stack_ip, stack_key)
+        except SoftTimeLimitExceeded:
+            raise
+        except Exception:
+            if was_resumed:
+                error_status = RESUME_FAILED
+                cleanup = CLEANUP_SUSPEND
             else:
-                ssh.close()
-                connected = True
+                error_status = CREATE_FAILED
+                cleanup = CLEANUP_DELETE
+
+            logger.error("Exception when checking SSH connection to stack "
+                         "[%s]: %s" % (self.stack_name,
+                                       traceback.format_exc()))
+            error_msg = ("Could not connect to your lab environment "
+                         "[%s]." % self.stack_name)
+            raise LaunchStackFailed(provider, error_status, error_msg,
+                                    cleanup)
+
+        return ssh
 
     def wait_for_rdp(self, stack_ip):
         port = getattr(self, 'port', None)
@@ -427,7 +519,7 @@ class LaunchStackTask(Task):
 
         connected = False
         s = socket.socket()
-        s.settimeout(self.sleep_seconds)
+        s.settimeout(self.get_sleep_timeout())
         while not connected:
             try:
                 s.connect((stack_ip, port))
@@ -461,10 +553,10 @@ class LaunchStackTask(Task):
 
         if stack_ip is None or not stack_key:
             if was_resumed:
-                error_status = 'RESUME_FAILED'
+                error_status = RESUME_FAILED
                 cleanup = CLEANUP_SUSPEND
             else:
-                error_status = 'CREATE_FAILED'
+                error_status = CREATE_FAILED
                 cleanup = CLEANUP_DELETE
 
             error_msg = ("Stack [%s] did not provide "
@@ -482,7 +574,26 @@ class LaunchStackTask(Task):
         # access to the training user while provisioning is going on.
         logger.info("Checking SSH connection "
                     "for stack [%s] at [%s]" % (self.stack_name, stack_ip))
-        self.wait_for_ssh(stack_key, stack_ip, was_resumed, provider)
+        ssh = self.wait_for_ssh(stack_key, stack_ip, was_resumed, provider)
+
+        try:
+            # If we're resuming and there's a resume hook, execute it.
+            if (was_resumed and
+                    self.hook_script and
+                    self.hook_events and
+                    isinstance(self.hook_events, dict) and
+                    self.hook_events.get("resume", False)):
+                logger.info("Executing post-resume hook for stack [%s] "
+                            "at [%s]" % (self.stack_name, stack_ip))
+                try:
+                    remote_exec(ssh, self.hook_script, params="resume")
+                except Exception as e:
+                    # We don't fail, as the user may have inadvertently broken
+                    # the stack.
+                    logger.error("Error running resume hook script on stack "
+                                 "[%s]: %s" % (self.stack_name, str(e)))
+        finally:
+            ssh.close()
 
         # If the protocol is RDP, wait for xrdp to come up.
         protocol = getattr(self, 'protocol', None)
@@ -500,7 +611,196 @@ class LaunchStackTask(Task):
         return check_data
 
 
-class CheckStudentProgressTask(Task):
+class SuspendStackTask(HastexoTask):
+    """
+    Suspends a stack.
+
+    """
+    def run(self, **kwargs):
+        for key, value in kwargs.items():
+            setattr(self, key, value)
+
+        stack = Stack.objects.get(id=self.stack_id)
+        self.stack_name = stack.name
+
+        try:
+            status = self.suspend_stack(stack)
+        except Exception as e:
+            error_msg = "Error suspending stack [%s]: %s" % (
+                self.stack_name,
+                str(e))
+            logger.error(error_msg)
+            status = SUSPEND_FAILED
+        else:
+            error_msg = ""
+
+        # The suspender doesn't check task results, so just save status in the
+        # database.
+        stack_data = {
+            'error_msg': error_msg,
+            'status': status,
+        }
+
+        self.update_stack_retry(stack_data)
+
+    def suspend_stack(self, stack):
+        provider = Provider.init(stack.provider)
+        provider_stack = provider.get_stack(stack.name)
+
+        if provider_stack["status"] in UP_STATES + (SUSPEND_FAILED,):
+            if (stack.hook_script and
+                    stack.hook_events and
+                    isinstance(stack.hook_events, dict) and
+                    stack.hook_events.get("suspend", False)):
+                try:
+                    # If there's a suspend hook, execute it.
+                    logger.info("Fetching pre-suspend hook [%s] for stack "
+                                "[%s]." % (stack.hook_script, stack.ip))
+                    script = read_from_contentstore(stack.course_id,
+                                                    stack.hook_script)
+                    logger.info("SSHing into stack [%s] at [%s]."
+                                % (stack.name, stack.ip))
+                    ssh = ssh_to(stack.user, stack.ip, stack.key)
+                    logger.info("Executing pre-suspend hook for stack [%s]."
+                                % stack.name)
+                    remote_exec(ssh, script, params="suspend")
+                except Exception as e:
+                    # We don't fail, as the user may have inadvertently broken
+                    # the stack.
+                    error_msg = str(e)
+                    logger.error("Error running pre-suspend hook for stack "
+                                 "[%s]: %s" % (stack.name, error_msg))
+                finally:
+                    try:
+                        ssh.close()
+                    except Exception:
+                        pass
+
+            # Suspend stack
+            logger.info("Suspending stack [%s]." % stack.name)
+            provider_stack = provider.suspend_stack(stack.name)
+        else:
+            logger.error("Cannot suspend stack with status [%s]." %
+                         provider_stack["status"])
+
+        return provider_stack["status"]
+
+
+class DeleteStackTask(HastexoTask):
+    """
+    Deletes a stack.
+
+    """
+    def run(self, **kwargs):
+        for key, value in kwargs.items():
+            setattr(self, key, value)
+
+        stack = Stack.objects.get(id=self.stack_id)
+        self.stack_name = stack.name
+
+        try:
+            status = self.delete_stack(stack)
+        except Exception as e:
+            status = DELETE_FAILED
+            provider = stack.provider
+            error_msg = "Error deleting stack [%s]: %s" % (
+                self.stack_name,
+                str(e))
+            logger.error(error_msg)
+        else:
+            provider = ""
+            error_msg = ""
+
+        # The reaper doesn't check task results, so just save status in the
+        # database.
+        stack_data = {
+            'status': status,
+            'provider': provider,
+            'error_msg': error_msg,
+        }
+        self.update_stack_retry(stack_data)
+
+    def delete_stack(self, stack):
+        """
+        Delete the stack.
+
+        """
+        settings = get_xblock_settings()
+        attempts = settings.get("delete_attempts", 3)
+        provider = Provider.init(stack.provider)
+        provider_stack = provider.get_stack(stack.name)
+        attempt = 0
+
+        while (provider_stack["status"] != DELETE_COMPLETE and
+               attempt < attempts):
+            if attempt:
+                self.sleep()
+
+            if provider_stack["status"] != DELETE_IN_PROGRESS:
+                logger.info("Attempt [%d] to delete stack [%s]." % (
+                    attempt, stack.name))
+
+                # Execute the pre-delete hook.
+                if (stack.hook_script and
+                        stack.hook_events and
+                        isinstance(stack.hook_events, dict) and
+                        stack.hook_events.get("delete", False)):
+                    try:
+                        # We need the stack to be up in order to execute the
+                        # pre-delete hook.
+                        if provider_stack["status"] == SUSPEND_COMPLETE:
+                            provider_stack = provider.resume_stack(
+                                stack.name)
+                        elif provider_stack["status"] not in UP_STATES:
+                            raise Exception("Invalid stack status [%s]." %
+                                            provider_stack["status"])
+                    except Exception as e:
+                        # We don't fail, as deleting the stack is more
+                        # important than running the pre-delete hook
+                        # successfully.
+                        logger.error("Could not resume stack [%s] for "
+                                     "pre-delete hook: %s" % (stack.name,
+                                                              str(e)))
+                    else:
+                        logger.info("Executing pre-delete hook for stack [%s] "
+                                    "at [%s]." % (stack.name, stack.ip))
+
+                        try:
+                            script = read_from_contentstore(stack.course_id,
+                                                            stack.hook_script)
+                            ssh = ssh_to(stack.user, stack.ip, stack.key)
+                            remote_exec(ssh, script, params="delete")
+                        except Exception as e:
+                            # Again, we don't fail, as deleting the stack is
+                            # paramount.
+                            logger.error("Could not execute pre-delete hook "
+                                         "for stack [%s]: %s" %
+                                         (stack.name, str(e)))
+                        finally:
+                            ssh.close()
+
+                try:
+                    provider_stack = provider.delete_stack(stack.name)
+                except SoftTimeLimitExceeded:
+                    # Retry on any exception except a timeout.
+                    raise
+                except Exception as e:
+                    logger.error("Failed to delete stack [%s]: %s" %
+                                 (stack.name, str(e)))
+                else:
+                    break
+
+            attempt += 1
+
+        if attempt == attempts:
+            error_msg = ("Failed to delete stack in [%d] attempts." %
+                         attempts)
+            raise Exception(error_msg)
+
+        return provider_stack["status"]
+
+
+class CheckStudentProgressTask(HastexoTask):
     """
     Check student progress by running a set of scripts via SSH.
 
@@ -516,11 +816,13 @@ class CheckStudentProgressTask(Task):
 
         try:
             # Open SSH connection to the public facing node
-            ssh = self.open_ssh_connection()
+            ssh = ssh_to(self.stack_user_name,
+                         self.stack_ip,
+                         self.stack_key)
 
             # Run tests on the stack
             res = self.run_tests(ssh)
-        except SoftTimeLimitExceeded:
+        except (RemoteExecTimeout, SoftTimeLimitExceeded):
             res = {
                 'status': 'CHECK_PROGRESS_TIMEOUT',
                 'error': True
@@ -538,27 +840,16 @@ class CheckStudentProgressTask(Task):
         score = 0
         errors = []
         for test in self.tests:
-            # Generate a temporary filename
-            script = '/tmp/.%s' % uuid.uuid4()
-
-            # Open the file remotely and write the script out to it.
-            f = sftp.open(script, 'w')
-            f.write(test)
-            f.close()
-
-            # Make it executable and run it.
-            sftp.chmod(script, 0o775)
-            _, stdout, stderr = ssh.exec_command(script)
-            retval = stdout.channel.recv_exit_status()
-            if retval == 0:
-                score += 1
+            try:
+                remote_exec(ssh, test, reuse_sftp=sftp)
+            except RemoteExecException as e:
+                error_msg = str(e)
+                if error_msg:
+                    errors.append(error_msg)
             else:
-                error = stderr.read()
-                if error:
-                    errors.append(error)
+                score += 1
 
-            # Remove the script
-            sftp.remove(script)
+        sftp.close()
 
         return {
             'status': 'CHECK_PROGRESS_COMPLETE',
@@ -566,11 +857,3 @@ class CheckStudentProgressTask(Task):
             'total': len(self.tests),
             'errors': errors
         }
-
-    def open_ssh_connection(self):
-        ssh = paramiko.SSHClient()
-        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
-        pkey = paramiko.RSAKey.from_private_key(StringIO(self.stack_key))
-        ssh.connect(self.stack_ip, username=self.stack_user_name, pkey=pkey)
-
-        return ssh
